@@ -88,6 +88,7 @@ ErrorStream <- R6Class(
   ),
   cloneable = FALSE
 )
+stdErrStream <- ErrorStream$new()
 
 #' @importFrom promises promise then finally is.promise %...>% %...!%
 rookCall <- function(func, req, data = NULL, dataLength = -1) {
@@ -102,7 +103,7 @@ rookCall <- function(func, req, data = NULL, dataLength = -1) {
 
     req$rook.input <- inputStream
 
-    req$rook.errors <- ErrorStream$new()
+    req$rook.errors <- stdErrStream
 
     req$httpuv.version <- httpuv_version()
 
@@ -230,7 +231,17 @@ AppWrapper <- R6Class(
       # The cpp_callback is an external pointer to a C++ function that writes
       # the response.
 
-      resp <- rookCall(private$app$call, req, req$.bodyData, seek(req$.bodyData))
+      resp <- if (is.null(private$app$call)) {
+        list(
+          status = 404L,
+          headers = list(
+            "Content-Type" = "text/plain"
+          ),
+          body = "404 Not Found\n"
+        )
+      } else {
+        rookCall(private$app$call, req, req$.bodyData, seek(req$.bodyData))
+      }
       # Note: rookCall() should never throw error because all the work is
       # wrapped in tryCatch().
 
@@ -276,7 +287,7 @@ AppWrapper <- R6Class(
     onWSClose = function(handle) {
       ws <- private$wsconns[[wsconn_address(handle)]]
       ws$handle <- NULL
-      private$wsconns[[wsconn_address(handle)]] <- NULL
+      rm(list = wsconn_address(handle), envir = private$wsconns)
 
       for (handler in ws$closeCallbacks) {
         handler()
@@ -341,6 +352,31 @@ AppWrapper <- R6Class(
 #'     }
 #'   }
 #'
+#' @examples
+#'
+#' \dontrun{
+#' # A WebSocket echo server that listens on port 8080
+#' startServer("0.0.0.0", 8080,
+#'   list(
+#'     onHeaders = function(req) {
+#'       # Print connection headers
+#'       cat(capture.output(str(as.list(req))), sep = "\n")
+#'     },
+#'     onWSOpen = function(ws) {
+#'       cat("Connection opened.\n")
+#'
+#'       ws$onMessage(function(binary, message) {
+#'         cat("Server received message:", message, "\n")
+#'         ws$send(message)
+#'       })
+#'       ws$onClose(function() {
+#'         cat("Connection closed.\n")
+#'       })
+#'
+#'     }
+#'   )
+#' )
+#' }
 #' @export
 WebSocket <- R6Class(
   'WebSocket',
@@ -401,10 +437,10 @@ WebSocket <- R6Class(
 #'   X, port numbers smaller than 1025 require root privileges.
 #' @param app A collection of functions that define your application. See
 #'   Details.
+#' @param quiet If \code{TRUE}, suppress error messages from starting app.
 #' @return A handle for this server that can be passed to
 #'   \code{\link{stopServer}} to shut the server down.
 #'
-
 #' @details \code{startServer} binds the specified port and listens for
 #'   connections on an thread running in the background. This background thread
 #'   handles the I/O, and when it receives a HTTP request, it will schedule a
@@ -525,8 +561,8 @@ WebSocket <- R6Class(
 #' s$stop()
 #' }
 #' @export
-startServer <- function(host, port, app) {
-  WebServer$new(host, port, app)
+startServer <- function(host, port, app, quiet = FALSE) {
+  WebServer$new(host, port, app, quiet)
 }
 
 #' @param name A string that indicates the path for the domain socket (on
@@ -539,8 +575,8 @@ startServer <- function(host, port, app) {
 #'   umask is left unchanged. (This parameter has no effect on Windows.)
 #' @rdname startServer
 #' @export
-startPipeServer <- function(name, mask, app) {
-  PipeServer$new(name, mask, app)
+startPipeServer <- function(name, mask, app, quiet = FALSE) {
+  PipeServer$new(name, mask, app, quiet)
 }
 
 #' Process requests
@@ -613,8 +649,10 @@ service <- function(timeoutMs = ifelse(interactive(), 100, 1000)) {
 #' If you have multiple hosts and/or ports to listen on, call the individual
 #' functions instead of \code{runServer}.
 #'
-#' @param host A string that is a valid IPv4 address that is owned by this
-#'   server, or \code{"0.0.0.0"} to listen on all IP addresses.
+#' @param host A string that is a valid IPv4 or IPv6 address that is owned by
+#'   this server, which the application will listen on. \code{"0.0.0.0"}
+#'   represents all IPv4 addresses and \code{"::/0"} represents all IPv6
+#'   addresses.
 #' @param port A number or integer that indicates the server port that should be
 #'   listened on. Note that on most Unix-like systems including Linux and Mac OS
 #'   X, port numbers smaller than 1025 require root privileges.
